using System;
using System.Collections.Generic;
using System.Text;

// This code is based on the following code project article by Rdiger Klaehn:
// http://www.codeproject.com/csharp/genericnumerics.asp

namespace MetaLabLogoProject
{

    /// <summary>
    /// Provides the interface for calculators that are used to provide
    /// generics with support for numbers.
    /// </summary>
    public interface ICalculator<T>
    {
        T Sum(T a, T b);
        T Mul(T a, T b);
        T Div(T a, T b);
        T Sub(T a, T b);
        T Sqrt(T a);
        T Pow(T a, double power);
        T Square(T a);
        int Compare(T a, T b);
        bool Less(T a, T b);
        bool LessOrEqual(T a, T b);
        bool Greater(T a, T b);
        bool GreaterOrEqual(T a, T b);
        bool Equal(T a, T b);
    }

    /// <summary>
    /// A generic Number class that can be used for generic mathematical
    /// primitives. This is necessary since .NET does not provide a way
    /// to restrict a template parameter to a number type and consequently
    /// does not allow to perform calculations with instances of does 
    /// template parameters.
    /// </summary>
    /// <typeparam name="T">The type you wish to use.</typeparam>
    /// <typeparam name="C">Calculator class for the given type.</typeparam>
    public struct Number<T, C> : IComparable<Number<T,C>>
        where C : ICalculator<T>,new()
    {
        private T value;
        private static C calculator = new C();

        private Number(T value)
        {
            this.value = value;
        }

        public Number<T, C> Square() { return calculator.Square(value); }
        public Number<T, C> Sqrt() { return calculator.Sqrt(value); }
        public Number<T, C> Pow(double power) { return calculator.Pow(value, power); }

        public static implicit operator Number<T, C>(T a) { return new Number<T, C>(a); }
        public static implicit operator T(Number<T, C> a) { return a.value; }

        public static Number<T, C> operator +(Number<T, C> a, Number<T, C> b) { return calculator.Sum(a.value, b.value); }
        public static Number<T, C> operator -(Number<T, C> a, Number<T, C> b) { return calculator.Sub(a.value, b.value); }
        public static Number<T, C> operator *(Number<T, C> a, Number<T, C> b) { return calculator.Mul(a.value, b.value); }
        public static Number<T, C> operator /(Number<T, C> a, Number<T, C> b) { return calculator.Div(a.value, b.value); }

        public static bool operator <(Number<T, C> a, Number<T, C> b) { return calculator.Less(a.value, b.value); }
        public static bool operator <=(Number<T, C> a, Number<T, C> b) { return calculator.LessOrEqual(a.value, b.value); }
        public static bool operator >(Number<T, C> a, Number<T, C> b) { return calculator.Greater(a.value, b.value); }
        public static bool operator >=(Number<T, C> a, Number<T, C> b) { return calculator.GreaterOrEqual(a.value, b.value); }
        public static bool operator ==(Number<T, C> a, Number<T, C> b) { return calculator.Equal(a.value, b.value); }
        public static bool operator !=(Number<T, C> a, Number<T, C> b) { return !(a == b); }

        public override int GetHashCode()
        {
            return value.GetHashCode();
        }

        public override bool Equals(object obj)
        {
            if (!(obj is Number<T, C>)) return false;

            return this == ((Number<T, C>)obj);
        }

        public int CompareTo(Number<T,C> other) { return calculator.Compare(value, other.value); }
    }

    /// <summary>
    /// A calculator implementation for double.
    /// </summary>
    public class DoubleCalculator : ICalculator<double>
    {
        public double Sum(double a, double b) { return a + b; }
        public double Mul(double a, double b) { return a * b; }
        public double Div(double a, double b) { return a / b; }
        public double Sub(double a, double b) { return a - b; }
        public double Sqrt(double a) { return Math.Sqrt(a); }
        public double Pow(double a, double power) { return Math.Pow(a, power); }
        public double Square(double a) { return a * a; }
        public int Compare(double a, double b) { return a.CompareTo(b); }
        public bool Less(double a, double b) { return a < b; }
        public bool LessOrEqual(double a, double b) { return a <= b; }
        public bool Greater(double a, double b) { return a > b; }
        public bool GreaterOrEqual(double a, double b) { return a >= b; }
        public bool Equal(double a, double b) { return a == b; }
    }
}
