using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;

using VectorStruct = MetaLabLogoProject.Vector3<double, MetaLabLogoProject.DoubleCalculator>;
using NumberType = System.Double;

namespace MetaLabLogoProject.GravityParticleSystem
{
    class GravityParticleSystem : IParticleSystem
    {
        static NumberType GravitationalConstant = 1;
        List<GravityParticle> particles = new List<GravityParticle>();
        double time = 0;

        public GravityParticleSystem()
        {
            Init();
        }

        public void Init()
        {
            Random rand = new Random();

            particles.Add(new GravityParticle(new Vector3<double,DoubleCalculator>(), new Vector3<double,DoubleCalculator>(),50,3000));
            for (int i = 0; i < 20; ++i)
            {
                    VectorStruct pos = new VectorStruct(rand.Next(-100,100), rand.Next(-100,100), rand.Next(-20,20));
                    pos.X += Math.Sign(pos.X) * 50;
                    pos.Y += Math.Sign(pos.Y) * 50;
                    pos.Z += Math.Sign(pos.Z) * 50;
                    VectorStruct velocity = new VectorStruct(pos.Y/50, -pos.X/50, 0);

                    particles.Add(new GravityParticle( pos, velocity, 10, 10));
            }
            //particles.Add(new GravityParticle(new Vector3<double,DoubleCalculator>(100,100,100),new Vector3<double,DoubleCalculator>(-2,-2,0), 10, 10));
            //particles.Add(new GravityParticle(new Vector3<double,DoubleCalculator>(-200,-100,-100), new Vector3<double, DoubleCalculator>(3,2, 0), 10, 10));
        }


        public void Step(double timeDelta)
        {
            //update velocities
            for (int i=0; i < particles.Count; ++i)
            {
                GravityParticle a = particles[i];
                for (int j=i+1; j < particles.Count; ++j)
                {
                    GravityParticle b = particles[j];

                    //get normalized direction and squared distance
                    VectorStruct directionToA = new VectorStruct(a.position);
                    directionToA.Sub(b.position);
                    NumberType distanceSquared = directionToA.LengthSquared;
                    directionToA.Normalize();

                    //calculate force
                    NumberType force = GravitationalConstant * a.mass * b.mass / distanceSquared;
                    force *= timeDelta; //integrate over time interval

                    //calculate velocity changes
                    VectorStruct directionToB = new VectorStruct(directionToA);
                    directionToA.ScalarMultiply(force / b.mass);
                    directionToB.ScalarMultiply(-force / a.mass);

                    //apply velocity changes
                    a.velocity.Add(directionToB);
                    b.velocity.Add(directionToA);
                }
            }

            //update positions and wobble
            foreach (GravityParticle particle in particles)
            {
                VectorStruct distanceTraveled = new VectorStruct(particle.velocity);
                distanceTraveled.ScalarMultiply(timeDelta);
                particle.position.Add(distanceTraveled);

                particle.wobbleRadians += timeDelta * particle.velocity.Length * 0.05;
            }

            time += timeDelta;
        }

        public void Render(System.Drawing.Graphics graphicsContext)
        {
            Random rand = new Random();
                foreach (GravityParticle particle in particles)
                {
                    double gray = Math.Min(Math.Max(particle.position.Z + 100, 0), 200);
                    double radius = gray * 0.125;

                    //wobbler
                    double radx, rady;
                    radx = rady = radius;

                    double wobble = Math.Sin(particle.wobbleRadians);
                    wobble *= wobble;
                    wobble /= 3;
                    wobble += 0.7;

                    //if (wobble > 0)
                    //{
                        //wobble += 0.7;
                        //wobble /= 1.7;

                        radx *= wobble;
                        //rady /= wobble;
                    //}
                        /*
                    else
                    {
                        wobble = -wobble;
                        wobble += 0.7;
                        wobble /= 1.7;

                        radx /= wobble;
                        rady *= wobble;
                        
                    }
                         * */

                    double epsilon = 0.001;

                    if (Math.Abs(particle.velocity.X) <= epsilon || Math.Abs(particle.velocity.Y) <= epsilon)
                    {
                        System.Diagnostics.Trace.WriteLine("border");
                    }

                    float angle = (float)Math.Atan2(particle.velocity.Y, particle.velocity.X);
                    angle = angle * 180 / (float)(Math.PI);

                    System.Drawing.Drawing2D.GraphicsState state;  

                    using (Pen p = new Pen(Color.FromArgb((int)gray + 55, (int)gray + 55, (int)gray + 55)))
                    {
                        state = graphicsContext.Save();
                        graphicsContext.TranslateTransform(
                            (float)(particle.position.X - radx),
                            (float)(particle.position.Y - rady));
                        graphicsContext.RotateTransform(angle);
                        graphicsContext.FillEllipse(
                            p.Brush,
                            (float)-radx,
                            (float)-radx,
                            (float)radx * 2,
                            (float)rady * 2
                            );
                        
                        /*graphicsContext.FillEllipse(
                            p.Brush,
                            (float)(particle.position.X - radx),
                            (float)(particle.position.Y - rady),
                            (float)radx * 2,
                            (float)rady * 2
                            );
                         * */
                        graphicsContext.RotateTransform(-angle);
                        graphicsContext.Restore(state);
                    }
                }
        }
    }
}
